from pathlib import PurePosixPath
from typing import Any, Dict
from copy import deepcopy
import fsspec

from kedro.io import AbstractVersionedDataSet, Version
from kedro.io.core import get_filepath_str, get_protocol_and_path


class PDFDataSet(AbstractVersionedDataSet[Any, Any]):
    """``PDFDataSet`` loads / save image data from a given filepath as `numpy` array using Pillow.

    Example:
    ::

        >>> PDFDataSet(filepath='/img/file/path.pdf')
    """

    DEFAULT_SAVE_ARGS = {}

    def __init__(self, filepath: str, save_args: Dict[str, Any] = None, version: Version = None, credentials: Dict[str, Any] = None, fs_args: Dict[str, Any] = None):
        """Creates a new instance of PDFDataSet to load / save image data for given filepath.

        Args:
            filepath: The location of the image file to load / save data.
        """
        _fs_args = deepcopy(fs_args) or {}
        _fs_open_args_load = _fs_args.pop("open_args_load", {})
        _fs_open_args_save = _fs_args.pop("open_args_save", {})
        _credentials = deepcopy(credentials) or {}

        protocol, path = get_protocol_and_path(filepath, version)

        self._protocol = protocol
        if protocol == "file":
            _fs_args.setdefault("auto_mkdir", True)
        self._fs = fsspec.filesystem(self._protocol, **_credentials, **_fs_args)

        super().__init__(
            filepath=PurePosixPath(path),
            version=version,
            exists_function=self._fs.exists,
            glob_function=self._fs.glob,
        )

        self._save_args = deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

        _fs_open_args_save.setdefault("mode", "wb")
        self._fs_open_args_load = _fs_open_args_load
        self._fs_open_args_save = _fs_open_args_save

    def _load(self) -> Any:
        """Loads data from the image file.

        Returns:
            Data from the image file as a numpy array
        """
        load_path = get_filepath_str(self._get_load_path(), self._protocol)
        with self._fs.open(load_path, **self._fs_open_args_load) as f:
            pdf = f.read()
            return pdf

    def _save(self, data: Any) -> None:
        """Saves image data to the specified filepath."""
        save_path = get_filepath_str(self._get_save_path(), self._protocol)
        with self._fs.open(save_path, **self._fs_open_args_save) as f:
            f.write(data)

    def _describe(self) -> Dict[str, Any]:
        """Returns a dict that describes the attributes of the dataset."""
        return dict(filepath=self._get_load_path(), protocol=self._protocol)


import re
from copy import deepcopy
from typing import Any, Dict, Union

import pandas as pd
from kedro.extras.datasets.pandas import ExcelDataSet
from kedro.io.core import PROTOCOL_DELIMITER, DataSetError, Version
from office365.runtime.auth.user_credential import UserCredential
from office365.sharepoint.files.file import File


class SharePointExcelDataSet(ExcelDataSet):
    def __init__(
        self,
        filepath: str,
        engine: str = "openpyxl",
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
        version: Version = None,
        credentials: Dict[str, Any] = None,
        fs_args: Dict[str, Any] = None,
    ) -> None:

        if not (
            credentials
            and "username" in credentials
            and credentials["username"]
            and "password" in credentials
            and credentials["password"]
        ):
            raise DataSetError(
                "'username' and 'password' argument cannot be empty. Please "
                "provide a Sharepoint credentials."
            )

        _credentials = deepcopy(credentials) or {}
        self._user_credentials = UserCredential(
            _credentials["username"], _credentials["password"]
        )

        super().__init__(
            filepath, engine, load_args, save_args, version, credentials, fs_args
        )

    def _load(self) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        load_path = str(self._get_load_path())
        if self._protocol == "file":
            # file:// protocol seems to misbehave on Windows
            # (<urlopen error file not on local host>),
            # so we don't join that back to the filepath;
            # storage_options also don't work with local paths
            return pd.read_excel(load_path, **self._load_args)

        load_path = f"{self._protocol}{PROTOCOL_DELIMITER}{load_path}"
        base_url = re.sub("(?<=com).*", "", load_path)

        relative_url = load_path.replace(base_url, "")
        file = (
            File.from_url(load_path)
            .with_credentials(self._user_credentials)
            .execute_query()
        )
        response = file.open_binary(file.context, relative_url)

        return pd.read_excel(
            response.content,
            **self._load_args
            # response.content, storage_options=self._storage_options, **self._load_args
        )



from .pdf_dataset import PDFDataSet
from .postgres_soft_replace_dataset import PostgresSoftReplaceDataSet
from .postgres_upsert_table import PostgresTableUpsertDataSet
from .sharepoint_excel_dataset import SharePointExcelDataSet

__all__ = [
    "PDFDataSet",
    "PostgresSoftReplaceDataSet",
    "PostgresTableUpsertDataSet",
    "SharePointExcelDataSet",
]



from typing import Any, Dict

import pandas as pd
from kedro.extras.datasets.pandas import SQLTableDataSet
from sqlalchemy.dialects.postgresql import insert
import sqlalchemy

class PostgresSoftReplaceFactory:
    def build(self):
        def postgres_upsert(table, conn, keys, dataframe_values):
            """
            Example
            -------
            test = pd.DataFrame({"name": ["Teste"]})

            replace_method_factory = PostgresSoftReplaceFactory()
            replace_method = replace_method_factory.build('roles_unique_name')
            df.to_sql('roles',
                engine,schema='public',
                if_exists='append',
                index=False,
                method=replace_method)

            See
            ----
            https://stackoverflow.com/questions/55187884/insert-into-postgresql-table-from-pandas-with-on-conflict-update
            https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method
            """
            delete_statement = sqlalchemy.delete(table.table)
            conn.execute(delete_statement)

            data = [dict(zip(keys, row)) for row in dataframe_values]
            insert_statement = insert(table.table).values(data)
            conn.execute(insert_statement)

        return postgres_upsert

class PostgresSoftReplaceDataSet(SQLTableDataSet):
    def __init__(
        self,
        table_name: str,
        credentials: Dict[str, Any],
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
    ) -> None:
        super().__init__(
            table_name, credentials, load_args, save_args
        )

    def _save(self, data: pd.DataFrame) -> None:
        replace_method_factory = PostgresSoftReplaceFactory()
        replace_method = replace_method_factory.build()
        engine = self.engines[self._connection_str]  # type: ignore
        data.to_sql(con=engine, **self._save_args, method=replace_method)
